from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from documented import DocumentedError
from octadocs.octiron import Octiron


@dataclass
class CacheNotFound(DocumentedError):
    """
    Cache file not found.

    File Path: {self.file_path}

    To create the cache, please run `mkdocs serve` or `mkdocs build`.
    """

    file_path: Path


@dataclass
class DiskCacheStorage:
    """Store the graph content on disk to support CLI queries."""

    octiron: Octiron
    directory: Optional[Path] = None
    format: str = 'json-ld'

    def _file_name(self) -> str:
        """Construct the file name."""
        return f'octadocs.{self.format}'

    def _file_path(self) -> Path:
        """Construct path to the file."""
        return (
            self.directory or
            self.octiron.root_directory.parent / '.cache'
        ) / self._file_name()

    def save(self) -> None:
        """Store graph contents on disk."""
        graph_content = self.octiron.graph.serialize(
            format=self.format,
        ).decode()

        try:
            self._file_path().write_text(graph_content)
        except FileNotFoundError:
            self._file_path().parent.mkdir(parents=True, exist_ok=True)
            self._file_path().write_text(graph_content)

    def load(self) -> None:
        """Load graph content from file."""
        try:
            graph_content = self._file_path().read_text()
        except FileNotFoundError:
            raise CacheNotFound(
                file_path=self._file_path(),
            )

        self.octiron.graph.parse(
            data=graph_content,
            format=self.format,
        )



import itertools

from octadocs.conversions import triples_to_quads
from octadocs.types import Triple
from owlrl import OWLRL_Extension
from owlrl.RDFS import rdf_type
from rdflib import BNode, Literal, Namespace, URIRef


class OWLRLExtensionNamedGraph(OWLRL_Extension):
    """OWL RL with inferred triples in a separate graph."""

    graph_name = URIRef('local:inference/owlrl')

    def flush_stored_triples(self) -> None:
        """Store the triples into a named graph."""
        # Remove the previous version of inferences
        self.graph.remove_context(self.graph_name)

        triples = itertools.starmap(
            Triple,
            self.added_triples,
        )
        quads = triples_to_quads(triples, graph=self.graph_name)
        self.graph.addN(quads)

        self.empty_stored_triples()

    def closure(self):
        """
        Generate the closure the graph. This is the real 'core'.

        The processing rules store new triples via the separate method :func:`.Core.store_triple` which stores
        them in the :code:`added_triples` array. If that array is empty at the end of a cycle,
        it means that the whole process can be stopped.

        If required, the relevant axiomatic triples are added to the graph before processing in cycles. Similarly
        the exchange of literals against bnodes is also done in this step (and restored after all cycles are over).
        """
        # Unfortunately, I had to copy-paste the content of `def closure()` from
        # the base class because I had to override the arguments of
        # `self.graph.add()`. See below.
        self.pre_process()

        # Handling the axiomatic triples. In general, this means adding all
        # tuples in the list that forwarded, and those include RDF or RDFS.
        # In both cases the relevant parts of the container axioms should also
        # be added.
        if self.axioms:
            self.add_axioms()

        # Add the datatype axioms, if needed (note that this makes use of the
        # literal proxies, the order of the call is important!
        if self.daxioms:
            self.add_d_axioms()

        self.flush_stored_triples()

        # Get first the 'one-time rules', ie, those that do not need an extra
        # round in cycles down the line
        self.one_time_rules()
        self.flush_stored_triples()

        # Go cyclically through all rules until no change happens
        new_cycle = True
        cycle_num = 0
        while new_cycle:
            # yes, there was a change, let us go again
            cycle_num += 1

            # DEBUG: print the cycle number out
            if self._debug:
                print("----- Cycle #:%d" % cycle_num)

            # go through all rules, and collect the replies (to see whether any
            # change has been done) the new triples to be added are collected
            # separately not to interfere with the current graph yet
            self.empty_stored_triples()

            # Execute all the rules; these might fill up the added triples array
            for t in self.graph:
                self.rules(t, cycle_num)

            # Add the tuples to the graph (if necessary, that is). If any new
            # triple has been generated, a new cycle will be necessary...
            new_cycle = len(self.added_triples) > 0

            for t in self.added_triples:
                # HERE IS THE PLACE where I changed the code. I just added the
                # graph qualifier.
                self.graph.add(t + (self.graph_name, ))

        self.post_process()
        self.flush_stored_triples()

        # Add possible error messages
        if self.error_messages:
            # I am not sure this is the right vocabulary to use for this
            # purpose, but I haven't found anything!
            # I could, of course, come up with my own, but I am not sure that
            # would be kosher...
            ERRNS = Namespace("http://www.daml.org/2002/03/agents/agent-ont#")
            self.graph.bind(
                "err",
                "http://www.daml.org/2002/03/agents/agent-ont#",
            )
            for m in self.error_messages:
                message = BNode()
                self.graph.add((message, rdf_type, ERRNS['ErrorMessage']))
                self.graph.add((message, ERRNS['error'], Literal(m)))



import csv
import sys
from pathlib import Path
from typing import Optional

from octadocs.cli.formatters.csv import csv_print
from octadocs.cli.formatters.json import print_json
from octadocs.cli.formatters.pretty import pretty_print
from octadocs.octiron import Octiron
from octadocs.storage import DiskCacheStorage
from octadocs.types import QueryResultsFormat, SelectResult
from typer import Argument, Option, Typer

app = Typer(name='sparql')


def create_octiron() -> Octiron:
    octiron = Octiron(root_directory=Path.cwd() / 'docs')

    disk_cache_storage = DiskCacheStorage(octiron=octiron)
    disk_cache_storage.load()

    return octiron


@app.callback(invoke_without_command=True)
def sparql(
    fmt: QueryResultsFormat = Option(
        default=QueryResultsFormat.PRETTY,
        metavar='format',
    ),
    query_text: Optional[str] = Argument(
        None,
        metavar='query',
        help='SPARQL query text. Will be read from stdin if empty.',
    ),
) -> None:
    """Run a SPARQL query against the graph."""
    if query_text is None:
        query_text = sys.stdin.read()

    octiron = create_octiron()

    query_result = octiron.query(query_text)

    {
        QueryResultsFormat.CSV: csv_print,
        QueryResultsFormat.PRETTY: pretty_print,
        QueryResultsFormat.JSON: print_json,
    }[fmt](query_result)   # type: ignore



from dominate.tags import a
from more_itertools import first
from rdflib import URIRef

from octadocs.octiron import Octiron


def link(octiron: Octiron, iri: URIRef) -> a:
    """Default facet to draw a link to something in HTML environment."""
    descriptions = octiron.query(
        '''
        SELECT * WHERE {
            ?page
                octa:url ?url ;
                octa:title ?label .
        } ORDER BY ?label LIMIT 1
        ''',
        page=iri,
    )
    location = first(descriptions, None)

    if not location:
        raise ValueError(f'Page not found by IRI: {iri}')

    return a(
        '📃 ',
        location['label'],
        href=location['url'],
    )



from itertools import starmap
from typing import Iterator

import rdflib
from octadocs.octiron.plugins import Loader
from octadocs.types import Triple


class TurtleLoader(Loader):
    """Read data from Turtle files."""

    regex = r'\.ttl$'

    def stream(self) -> Iterator[Triple]:
        """Read a Turtle file and return a stream of triples."""
        graph = rdflib.Graph()
        graph.parse(source=str(self.path), format='turtle')
        return starmap(Triple, iter(graph))



import datetime
import json
from dataclasses import dataclass
from functools import partial
from itertools import starmap
from pprint import pformat
from typing import Any, Dict, Iterator, List, TypeVar, Union

import rdflib
from boltons.iterutils import remap
from documented import DocumentedError
from octadocs.octiron.context import merge
from octadocs.pyld_document_loaders.choice import ChoiceLoader
from octadocs.pyld_document_loaders.named import NamedLoader
from octadocs.types import LOCAL, Context, Triple
from pyld import jsonld
from pyld.documentloader.requests import requests_document_loader

try:  # noqa
    from yaml import CSafeDumper as SafeDumper  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeDumper  # type: ignore   # noqa
    from yaml import SafeLoader  # type: ignore   # noqa


MetaData = Union[List[Dict[str, Any]], Dict[str, Any]]   # type: ignore  # noqa

Data = TypeVar('Data')


@dataclass
class ExpandError(DocumentedError):
    """
    JSON-LD expand operation failed. Please review the source data below.

    {self.formatted_data}
    """

    meta_data: MetaData

    @property
    def formatted_data(self) -> str:
        """Format meta_data for printing."""
        return pformat(self.meta_data, indent=4)


def _convert(term: Any) -> Any:  # type: ignore
    """Convert $statement to @statement."""
    if isinstance(term, str) and term.startswith('$'):
        return '@' + term[1:]  # noqa: WPS336

    # pyld cannot expand() a document which contains data which cannot be
    # trivially serialized to JSON.
    # FIXME we may want to replace this with an xsd:date declaration, but let's
    #   revisit that later.
    if isinstance(term, (datetime.date, datetime.datetime)):
        return str(term)

    return term


def convert_dollar_signs(
    meta_data: Data,
) -> Data:
    """
    Convert $ character to @ in keys.

    We use $ by convention to avoid writing quotes.
    """
    if not meta_data:
        return {}

    return remap(
        meta_data,
        lambda path, key, value: (  # noqa: WPS110
            _convert(key),
            _convert(value),
        ),
    )


def as_triple_stream(
    raw_data: MetaData,
    context: Context,
    local_iri: str,
    named_contexts: Dict[str, Any],
) -> Iterator[Triple]:
    """Convert YAML dict to a stream of triples."""
    if isinstance(raw_data, list):
        yield from _list_as_triple_stream(
            context=context,
            local_iri=local_iri,
            raw_data=raw_data,
            named_contexts=named_contexts,
        )

    elif isinstance(raw_data, dict):
        yield from _dict_as_triple_stream(
            context=context,
            local_iri=local_iri,
            raw_data=raw_data,
            named_contexts=named_contexts,
        )

    else:
        raise ValueError(f'Format of data not recognized: {raw_data}')


def reformat_blank_nodes(prefix: str, triple: Triple) -> Triple:
    """
    Prepend prefix to every blank node.

    JSON-LD flattening creates sequential identifiers in the form of _:bN
    for every blank node. See https://www.w3.org/TR/json-ld11/

    This means that subgraphs of the Octiron ConjunctiveGraph have clashing
    blank node identifiers.

    To avoid that, we prepend a prefix to every blank node identifier.
    """
    return Triple(*(
        rdflib.BNode(value=prefix + str(singleton)) if (
            isinstance(singleton, rdflib.BNode)
        ) else singleton
        for singleton in triple
    ))


def _dict_as_triple_stream(  # type: ignore
    raw_data: Dict[str, Any],
    context: Context,
    local_iri: str,
    named_contexts: Dict[str, Any],
) -> Iterator[Triple]:
    """Convert dict into a triple stream."""
    meta_data = convert_dollar_signs(raw_data)

    local_context = meta_data.pop('@context', None)
    if isinstance(local_context, str):
        try:
            local_context = named_contexts[local_context]
        except KeyError as err:
            raise ValueError(
                f'Named context {local_context} is unknown.',
            ) from err

    if local_context is not None:
        context = merge(
            first=context,
            second=local_context,
        )

    if meta_data.get('@id') is None:
        # The document author did not tell us about what their document is.
        # In this case, we assume that the local_iri of the document file
        # is the subject of the document description.
        meta_data['@id'] = local_iri

    # The author specified an IRI the document tells us about. Let us
    # link this IRI to the local document IRI.
    meta_data['octa:subjectOf'] = local_iri

    # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/97
    # If we don't expand with an explicit @base, import will fail silently.
    try:
        meta_data = jsonld.expand(
            meta_data,
            options={
                'base': str(LOCAL),
                'expandContext': context,
                'documentLoader': ChoiceLoader(loaders=[
                     NamedLoader(documents=named_contexts),
                     requests_document_loader(),
                ]),
            },
        )
    except TypeError as err:
        raise ExpandError(
            meta_data=meta_data,
        ) from err

    # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/98
    # If we don't flatten, @included sections will not be imported.
    meta_data = jsonld.flatten(meta_data)
    serialized_meta_data = json.dumps(meta_data, indent=4)

    graph = rdflib.Graph()
    graph.parse(
        data=serialized_meta_data,
        format='json-ld',
    )
    yield from map(
        partial(
            reformat_blank_nodes,
            f'{local_iri}/',
        ),
        starmap(
            Triple,
            iter(graph),
        ),
    )


def _list_as_triple_stream(  # type: ignore
    raw_data: List[Dict[str, Any]],
    context: Context,
    local_iri: str,
    named_contexts: Dict[str, Any],
) -> Iterator[Triple]:
    """Convert a list into a triple stream."""
    for sub_document in raw_data:
        yield from as_triple_stream(
            raw_data=sub_document,
            context=context,
            local_iri=local_iri,
            named_contexts=named_contexts,
        )



"""
Iolanta facet management.

This module contains a few functions which later will be refactored into
Iolanta - the generic metaverse/cyberspace browser.
"""
import operator
import pydoc
from dataclasses import dataclass
from typing import Protocol, Optional, Callable, Union

from documented import DocumentedError
from more_itertools import first
from rdflib.term import Node, URIRef
from urlpath import URL

from octadocs.octiron import Octiron


HTML = URIRef('https://html.spec.whatwg.org/')


@dataclass
class FacetNotCallable(DocumentedError):
    """
    Python facet not callable.

      - Import path: {self.path}
      - Object imported: {self.facet}

    The imported Python object is not a callable and thus cannot be used as a
    facet.
    """

    path: str
    facet: object


@dataclass
class FacetNotFound(DocumentedError):
    """
    Facet not found.

        Node: {self.node}
        Environment: {self.environment}

    We could not find a facet to display this node 😟
    """

    node: Node
    environment: URIRef


class PythonFacet(Protocol):
    def __call__(
        self,
        octiron: Octiron,
        node: Node,
        environment: URIRef,
    ) -> str:
        ...


def find_default_facet_iri_for_environment(octiron: Octiron, environment: URIRef):
    rows = octiron.query(
        '''
        SELECT * WHERE {
            ?environment iolanta:hasDefaultFacet ?facet .
        }
        ''',
        environment=environment,
    )

    facets = map(operator.itemgetter('facet'), rows)

    return first(facets, None)


def find_facet_iri(
    octiron: Octiron,
    environment: URIRef,
    node: Node,
) -> Optional[URIRef]:
    if not isinstance(node, URIRef):
        node = URIRef(node)

    rows = octiron.query(
        '''
        SELECT ?facet WHERE {
            ?node iolanta:facet ?facet .
            ?facet iolanta:supports ?environment .
        }
        ''',
        node=node,
        environment=environment,
    )

    facets = map(operator.itemgetter('facet'), rows)

    if facet := first(facets, None):
        return facet

    facet = find_default_facet_iri_for_environment(
        octiron=octiron,
        environment=environment,
    )

    if facet is None:
        raise FacetNotFound(
            node=node,
            environment=environment,
        )

    return facet


def resolve_facet(iri: URIRef) -> Callable[[Octiron, Node], str]:
    url = URL(str(iri))

    if url.scheme != 'python':
        raise Exception(
            'Octadocs only supports facets which are importable Python '
            'callables. The URLs of such facets must start with `python://`, '
            'which {url} does not comply to.'.format(
                url=url,
            )
        )

    facet = pydoc.locate(url.hostname)

    if not callable(facet):
        raise FacetNotCallable(
            path=url,
            facet=facet,
        )

    return facet


def render(
    node: Union[str, Node],
    octiron: Octiron,
    environment: URIRef = HTML,
) -> str:
    """Find an Iolanta facet for a node and render it."""
    facet_iri = find_facet_iri(
        octiron=octiron,
        environment=environment,
        node=node,
    )

    facet = resolve_facet(iri=facet_iri)

    return facet(
        octiron=octiron,
        iri=node,
    )
